/* $Id: bitmap.c,v 1.9 2010/03/01 23:55:14 kiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2010  Robert Kiesling, ctalk@ctalklang.org.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

#if ! defined (DJGPP) && ! defined (WITHOUT_X11)
#include <Xlib.h>
#include "x11defs.h"

extern Display *display;

#define RECTANGLE_GCV_MASK (GCFunction|GCForeground|GCFillStyle) 
int __xlib_clear_pixmap (Drawable d, unsigned int width, 
			 unsigned int height, GC gc){
  XGCValues rectangle_gcv, old_gcv;
  XGetGCValues (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  rectangle_gcv.function = GXcopy;
  rectangle_gcv.foreground = WhitePixel (display, DefaultScreen (display));
  rectangle_gcv.fill_style = FillSolid;
  XChangeGC (display, gc, RECTANGLE_GCV_MASK, &rectangle_gcv);
  XFillRectangle (display, d, gc,
		  0, 0, width, height);
  XChangeGC (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  return SUCCESS;
}

int __xlib_clear_rectangle_basic (Drawable d, int x, int y, 
				  unsigned int width, unsigned int height,
				  GC gc) {
  XGCValues rectangle_gcv, old_gcv;
  XGetGCValues (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  rectangle_gcv.function = GXset;
  rectangle_gcv.foreground = WhitePixel (display, DefaultScreen (display));
  rectangle_gcv.fill_style = FillSolid;
  XChangeGC (display, gc, RECTANGLE_GCV_MASK, &rectangle_gcv);
  XFillRectangle (display, d, gc,
 		  x, y, width, height);
  XChangeGC (display, gc, DEFAULT_GCV_MASK, &old_gcv);
  return SUCCESS;
}

int __ctalkX11CreatePaneBuffer (OBJECT *pane,
				int width, int height, 
				int depth) {
  OBJECT *pane_object, *paneBuffer_object, *paneBackingStore_object,
    *xWindowID_object;
  OBJECT *xGC_object;
  GC gc;
  int win_id;
  char buf[MAXMSG];
  Pixmap p;
  pane_object = (IS_VALUE_INSTANCE_VAR(pane) ? pane -> __o_p_obj :
		 pane);
  if ((xGC_object = 
	__ctalkGetInstanceVariable (pane_object, "xGC", TRUE))
       == NULL)
    return ERROR;
  sscanf (xGC_object->instancevars->__o_value, "0x%p", &gc);
  if ((paneBuffer_object = 
       __ctalkGetInstanceVariable (pane_object, "paneBuffer", TRUE))
      == NULL)
    return ERROR;
  if ((xWindowID_object = 
       __ctalkGetInstanceVariable (pane_object, "xWindowID", TRUE))
      == NULL)
    return ERROR;
  if ((paneBackingStore_object = 
	__ctalkGetInstanceVariable (pane_object, "paneBackingStore", TRUE))
       == NULL)
    return ERROR;
  win_id = atoi (xWindowID_object->instancevars->__o_value);
  p = XCreatePixmap (display, win_id, width, height, depth);
  __xlib_clear_pixmap (p, width, height, gc);
  sprintf (buf, "%d", (int)p);
  __ctalkSetObjectValueVar (paneBuffer_object, buf);
  p = XCreatePixmap (display, win_id, width, height, depth);
  sprintf (buf, "%d", (int)p);
  __ctalkSetObjectValueVar (paneBackingStore_object, buf);
  return SUCCESS;
}

int __ctalkX11FreePaneBuffer (OBJECT *pane) {
  OBJECT *pane_object, *paneBuffer_object, *paneBackingStore_object;
  pane_object = (IS_VALUE_INSTANCE_VAR(pane) ? pane -> __o_p_obj :
		 pane);
  if ((paneBuffer_object = 
       __ctalkGetInstanceVariable (pane_object, "paneBuffer", TRUE))
      == NULL)
    return ERROR;
  if ((paneBackingStore_object = 
	__ctalkGetInstanceVariable (pane_object, "paneBackingStore", TRUE))
       == NULL)
    return ERROR;
  XFreePixmap (display, 
	       atoi(paneBackingStore_object->instancevars->__o_value));
  XFreePixmap (display, 
	       atoi(paneBuffer_object->instancevars->__o_value));
  return SUCCESS;
}
int __ctalkX11ResizePaneBuffer (OBJECT *pane,
				int width, int height) {
  OBJECT *pane_object, *paneBuffer_object, *paneBackingStore_object,
    *xWindowID_object;
  OBJECT *xGC_object;
  int win_id;
  char buf[MAXMSG];
/*   Pixmap p, p_old; */
  unsigned int p_old;
  Pixmap p;
  GC gc;
  pane_object = (IS_VALUE_INSTANCE_VAR(pane) ? pane -> __o_p_obj :
		 pane);
  if ((paneBuffer_object = 
       __ctalkGetInstanceVariable (pane_object, "paneBuffer", TRUE))
      == NULL)
    return ERROR;
  if ((xWindowID_object = 
       __ctalkGetInstanceVariable (pane_object, "xWindowID", TRUE))
      == NULL)
    return ERROR;
  if ((paneBackingStore_object = 
	__ctalkGetInstanceVariable (pane_object, "paneBackingStore", TRUE))
       == NULL)
    return ERROR;
  if ((xGC_object = 
	__ctalkGetInstanceVariable (pane_object, "xGC", TRUE))
       == NULL)
    return ERROR;
  win_id = atoi (xWindowID_object->instancevars->__o_value);
  p = XCreatePixmap (display, win_id, width, height, CopyFromParent);
  sscanf (paneBuffer_object->instancevars->__o_value, "%u", &p_old);
  sscanf (xGC_object->instancevars->__o_value, "0x%p", &gc);
  XCopyArea (display, (Pixmap)p_old, p, gc, 
	     0, 0, 
	     width, height,
	     0, 0);
  sprintf (buf, "%u", (unsigned int)p);
  __ctalkSetObjectValueVar (paneBuffer_object, buf);
  XFreePixmap (display, (Pixmap)p_old);
  sscanf (paneBackingStore_object->instancevars->__o_value, "%u", &p_old);
  p = XCreatePixmap (display, win_id, width, height, CopyFromParent);
  XCopyArea (display, (Pixmap)p_old, p, gc, 
	     0, 0, 
	     width, height,
	     0, 0);
  sprintf (buf, "%u", (unsigned int)p);
  __ctalkSetObjectValueVar (paneBackingStore_object, buf);
  XFreePixmap (display, (Pixmap)p_old);
  return SUCCESS;
}

int __ctalkX11ClearBufferRectangle (OBJECT *pane,
				    int x, int y,
				    int width, int height) {
  OBJECT *pane_object, *paneBuffer_object, *paneBackingStore_object,
    *xWindowID_object, *xGC_object;
  GC gc;
  pane_object = (IS_VALUE_INSTANCE_VAR(pane) ? pane -> __o_p_obj :
		 pane);
  if ((xGC_object = 
       __ctalkGetInstanceVariable (pane_object, "xGC", TRUE))
      == NULL)
    return ERROR;
  if ((paneBuffer_object = 
       __ctalkGetInstanceVariable (pane_object, "paneBuffer", TRUE))
      == NULL)
    return ERROR;
  sscanf (xGC_object->instancevars->__o_value, "0x%p", &gc);
  __xlib_clear_rectangle_basic 
    (atoi(paneBuffer_object->instancevars->__o_value), 
     x, y, width, height, gc);
  if ((paneBackingStore_object = 
	__ctalkGetInstanceVariable (pane_object, "paneBackingStore", TRUE))
       == NULL)
    return ERROR;
  __xlib_clear_rectangle_basic 
    (atoi(paneBackingStore_object->instancevars->__o_value), 
     x, y, width, height, gc);
  if ((xWindowID_object = 
       __ctalkGetInstanceVariable (pane_object, "xWindowID", TRUE))
      == NULL)
    return ERROR;
  __xlib_clear_rectangle_basic 
    (atoi(xWindowID_object->instancevars->__o_value), 
     x, y, width, height, gc);
  return SUCCESS;

}

#else

int __ctalkX11CreatePaneBuffer (OBJECT *pane_object,
				  int width, int height, int depth) {
  return SUCCESS;
}

int __ctalkX11FreePaneBuffer (OBJECT *pane_object) {
  return SUCCESS;
}

int __ctalkX11ResizePaneBuffer (OBJECT *pane_object,
				int width, int height) {
  return SUCCESS;
}

int __ctalkX11ClearBufferRectangle (OBJECT *pane_object,
				    int x, int y,
				    int width, int height) {
  return SUCCESS;
}

#endif /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
