/* $Id: font.c,v 1.3 2010/02/19 11:44:36 kiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2010  Robert Kiesling, ctalk@ctalklang.org.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

#if ! defined (DJGPP) && ! defined (WITHOUT_X11)
#include <Xlib.h>

char *__get_actual_xlfd (char *xlfd) {
  Display *display;
  char *d_env = NULL;
  char **font_set = NULL;
  static char actual_xlfd[MAXMSG];
  int n_fonts = 0;
  if ((display = XOpenDisplay (d_env)) == NULL) {
    _warning ("__get_actual_xlfd: Could not open display %s.\n",
 	      getenv ("DISPLAY"));
    return NULL;
  }
  font_set = XListFonts (display, xlfd, 10, &n_fonts);
  XCloseDisplay (display);
  if (n_fonts) {
    strcpy (actual_xlfd, font_set[0]);
    XFreeFontNames (font_set);
    return actual_xlfd;
  }
  return NULL;
}

int __ctalkX11QueryFont (OBJECT *font, char *xlfd) {
  Display *display;
  char *d_env = NULL;
  char **font_set;
  char buf[MAXMSG];
  int n_fonts;
  XFontStruct *font_info;
  OBJECT *font_object, *fontId_object, *ascent_object, *descent_object,
    *maxWidth_object, *height_object, *fontDesc_object;
  font_object = (IS_VALUE_INSTANCE_VAR (font) ? font->__o_p_obj : font);
  if ((display = XOpenDisplay (d_env)) == NULL) {
    _warning ("__ctalkX11QueryFont: Could not open display %s.\n",
  	      getenv ("DISPLAY"));
    return ERROR;
  }
  font_set = XListFonts (display, xlfd, 10, &n_fonts);
  if (n_fonts) {
    font_info = XLoadQueryFont (display, xlfd);
    if ((fontId_object = 
	 __ctalkGetInstanceVariable (font_object, "fontId", TRUE)) != NULL) {
      sprintf (buf, "%d", (int) font_info -> fid);
      __ctalkSetObjectValue (fontId_object, buf);
      __ctalkSetObjectValueVar (fontId_object, buf);
    }
    if ((ascent_object = 
	 __ctalkGetInstanceVariable (font_object, "ascent", TRUE)) != NULL) {
      sprintf (buf, "%d", font_info -> ascent);
      __ctalkSetObjectValue (ascent_object, buf);
      __ctalkSetObjectValueVar (ascent_object, buf);
    }
    if ((descent_object = 
	 __ctalkGetInstanceVariable (font_object, "descent", TRUE)) != NULL) {
      sprintf (buf, "%d", font_info -> descent);
      __ctalkSetObjectValue (descent_object, buf);
      __ctalkSetObjectValueVar (descent_object, buf);
    }
    if ((maxWidth_object = 
	 __ctalkGetInstanceVariable (font_object, "maxWidth", TRUE)) != NULL) {
      sprintf (buf, "%d", 
	       (font_info -> max_bounds.rbearing - 
		font_info -> min_bounds.lbearing));
      __ctalkSetObjectValue (maxWidth_object, buf);
      __ctalkSetObjectValueVar (maxWidth_object, buf);
    }
    if ((height_object = 
	 __ctalkGetInstanceVariable (font_object, "height", TRUE)) != NULL) {
      sprintf (buf, "%d", 
	       (font_info -> max_bounds.ascent + 
		font_info -> max_bounds.descent));
      __ctalkSetObjectValue (height_object, buf);
      __ctalkSetObjectValueVar (height_object, buf);
    }
    if ((fontDesc_object = 
	 __ctalkGetInstanceVariable (font_object, "fontDesc", TRUE)) != NULL){
      __ctalkSetObjectValue (fontDesc_object, font_set[0]);
      __ctalkSetObjectValueVar (fontDesc_object, font_set[0]);
    }
    XFreeFontNames(font_set);
    XFreeFont (display, font_info);
  } else {
    _warning ("__ctalkX11QueryFont: Could not find font %s.\n", xlfd);
  }
  XCloseDisplay (display);
  return SUCCESS;
}

#else
char *__get_actual_xlfd (char *xlfd) {
  return SUCCESS;
}

int __ctalkX11QueryFont (OBJECT *font, char *xlfd) {
  return SUCCESS;
}

#endif /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
