/* $Id: guidrawline.c,v 1.1 2010/03/13 20:01:31 kiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2010  Robert Kiesling, ctalk@ctalklang.org.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 *  __ctalkX11PaneDrawPoint is here for backward compatibility but is
 *  deprecated.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

#if ! defined (DJGPP) && ! defined (WITHOUT_X11)
#include "x11defs.h"

extern Display *display;   /* Defined in x11lib.c. */
extern char *shm_mem;
extern int mem_id;

#if X11LIB_FRAME
int __ctalkX11PaneDrawLine (OBJECT *self, OBJECT *line, OBJECT *pen) {
  return SUCCESS;
}
int __ctalkGUIPaneDrawLine (OBJECT *self, OBJECT *line, OBJECT *pen) {
  return SUCCESS;
}
#else /* X11LIB_FRAME */
int __ctalkX11PaneDrawLine (OBJECT *self, OBJECT *line, OBJECT *pen) {
  return __ctalkGUIPaneDrawLine (self, line, pen);
}

int __ctalkGUIPaneDrawLine (OBJECT *self, OBJECT *line, OBJECT *pen) {
  OBJECT *self_object, *pen_object; 
  OBJECT *line_start_var, *line_start_x_var, *line_start_y_var;
  OBJECT *line_end_var, *line_end_x_var, *line_end_y_var;
  OBJECT *pen_width_object, *pen_color_object; 
  OBJECT *win_id_value, *gc_value;
  OBJECT *line_object;
  OBJECT *paneBuffer_object, *paneBackingStore_object;
  char colorname[MAXMSG];
  char d_buf[MAXMSG];
#ifdef GRAPHICS_WRITE_SEND_EVENT
  XEvent send_event;
#endif

  self_object = (IS_VALUE_INSTANCE_VAR (self) ? self->__o_p_obj : self);
  line_object = (IS_VALUE_INSTANCE_VAR (line) ? line->__o_p_obj : line);
  win_id_value = __x11_pane_win_id_value_object (self_object);
  gc_value = __x11_pane_win_gc_value_object (self_object);
  pen_object = (IS_VALUE_INSTANCE_VAR (pen) ? pen->__o_p_obj : pen);
  pen_width_object = 
    __ctalkGetInstanceVariable (pen_object, "width", TRUE);
  pen_color_object = 
    __ctalkGetInstanceVariable (pen_object, "colorName", TRUE);
  strcpy (colorname, pen_color_object->instancevars->__o_value);
  line_start_var = 
    __ctalkGetInstanceVariable (line_object, "start", TRUE);
  line_start_x_var = 
    __ctalkGetInstanceVariable (line_start_var, "x", TRUE);
  line_start_y_var = 
    __ctalkGetInstanceVariable (line_start_var, "y", TRUE);
  line_end_var = 
    __ctalkGetInstanceVariable (line_object, "end", TRUE);
  line_end_x_var = 
    __ctalkGetInstanceVariable (line_end_var, "x", TRUE);
  line_end_y_var = 
    __ctalkGetInstanceVariable (line_end_var, "y", TRUE);
  if (__have_bitmap_buffers (self_object)) {
    if ((paneBuffer_object =
	 __ctalkGetInstanceVariable (self_object, "paneBuffer", TRUE))
	== NULL) 
      return ERROR;
    if ((paneBackingStore_object =
	 __ctalkGetInstanceVariable (self_object, "paneBackingStore", TRUE))
	== NULL) 
      return ERROR;
    sprintf (d_buf, "%d:%d:%d:%d:%d:%d:%d:%s",
	     atoi (line_start_x_var->instancevars->__o_value),
	     atoi (line_start_y_var->instancevars->__o_value),
	     atoi (line_end_x_var->instancevars->__o_value),
	     atoi (line_end_y_var->instancevars->__o_value),
	     atoi (pen_width_object->instancevars->__o_value),
	     atoi (paneBuffer_object->instancevars->__o_value),
	     atoi (paneBackingStore_object->instancevars->__o_value),
	     colorname );
  } else {
    sprintf (d_buf, "%d:%d:%d:%d:%d:%d:%d:%s",
	     atoi (line_start_x_var->instancevars->__o_value),
	     atoi (line_start_y_var->instancevars->__o_value),
	     atoi (line_end_x_var->instancevars->__o_value),
	     atoi (line_end_y_var->instancevars->__o_value),
	     atoi (pen_width_object->instancevars->__o_value),
	     0, 0,
	     colorname );
  }
  make_req (shm_mem, PANE_DRAW_LINE_REQUEST,
	    win_id_value->__o_value,
	    gc_value->__o_value, d_buf);
#ifdef GRAPHICS_WRITE_SEND_EVENT
  send_event.xgraphicsexpose.type = GraphicsExpose;
  send_event.xgraphicsexpose.send_event = True;
  send_event.xgraphicsexpose.display = display;
  send_event.xgraphicsexpose.drawable = atoi(win_id_value->__o_value);
  XSendEvent (display, atoi(win_id_value->__o_value),
 	      False, 0L, &send_event);
#endif
  wait_req (shm_mem);
  return SUCCESS;
}

#endif /* X11LIB_FRAME */
#else /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
static void __gui_support_error (void) {
  _error ("__ctalkGUI ... () functions require Graphical User Interface support.\n");
}
int __ctalkX11PaneDrawLine (OBJECT *self, OBJECT *line, OBJECT *pen) {__gui_support_error(); return ERROR;}
int __ctalkGUIPaneDrawLine (OBJECT *self, OBJECT *line, OBJECT *pen) {__gui_support_error(); return ERROR;}
#endif /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
