/* $Id: guiputstr.c,v 1.2 2010/03/16 20:00:25 kiesling Exp $ -*-c-*-*/

/*
  This file is part of Ctalk.
  Copyright  2005-2010  Robert Kiesling, ctalk@ctalklang.org.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 *  __ctalkX11PanePutStr is here for compatibility,
 *  but it is deprecated.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"

#if ! defined (DJGPP) && ! defined (WITHOUT_X11)
#include "x11defs.h"

extern Display *display;   /* Defined in x11lib.c. */
extern char *shm_mem;
extern int mem_id;
extern char *fixed_font_name;
extern Font fixed_font;

#if X11LIB_FRAME
int __ctalkX11PanePutStr (OBJECT *self_object, int x, int y, char *s) {
  return SUCCESS;
}
int __ctalkGUIPanePutStr (OBJECT *self_object, int x, int y, char *s) {
  return SUCCESS;
}
#else /* X11LIB_FRAME */

OBJECT *__x11_pane_font_id_value_object (OBJECT *);

int __ctalkX11PanePutStr (OBJECT *self, int x, int y, char *s) {
  return __ctalkGUIPanePutStr (self, x, y, s);
}
int __ctalkGUIPanePutStr (OBJECT *self, int x, int y, char *s) {
  OBJECT *self_object, *win_id_value, *gc_value, *font_id_value;
  OBJECT *fontDesc_var, *fontDesc_value_var;
  OBJECT *paneBuffer_var, *paneBackingStore_var;
  int pane_width;
  char d_buf[MAXLABEL], f_buf[MAXLABEL], fname_buf[MAXLABEL];
#ifdef GRAPHICS_WRITE_SEND_EVENT
  XEvent send_event;
#endif
  self_object = (IS_VALUE_INSTANCE_VAR (self) ? self->__o_p_obj : self);
  win_id_value = __x11_pane_win_id_value_object (self_object);
  font_id_value = __x11_pane_font_id_value_object (self_object);
  pane_width = __pane_x_size (self_object);
  if ((fontDesc_var = 
       __ctalkGetInstanceVariable (self_object, "fontDesc", TRUE)) 
      == NULL)
    return 0;
  if ((fontDesc_value_var = 
       __ctalkGetInstanceVariable (fontDesc_var, "value", TRUE)) 
      == NULL)
    return 0;
  gc_value = __x11_pane_win_gc_value_object (self_object);
  if (atoi (font_id_value -> __o_value) == 0) {
    sprintf (f_buf, "%d", (int)fixed_font);
    strcpy (fname_buf, fixed_font_name);
  } else {
    strcpy (f_buf, font_id_value -> __o_value);
    strcpy (fname_buf, fontDesc_value_var -> __o_value);
  }
  sprintf (d_buf, "%d:%d:%d:%s:%s:%s", x, y, pane_width,
	   f_buf, 
	   fname_buf,
	   s);
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO    
  fprintf (stderr, "__ctalkX11LibPutStr: d_buf  = %s.\n", d_buf);
#endif
  if (__have_bitmap_buffers (self_object)) {
    if ((paneBuffer_var =
	 __ctalkGetInstanceVariable (self_object, "paneBuffer", TRUE))
	== NULL) 
      return ERROR;
    if ((paneBackingStore_var =
	 __ctalkGetInstanceVariable (self_object, "paneBackingStore", TRUE))
	== NULL) 
      return ERROR;
    make_req (shm_mem, PANE_PUT_STR_REQUEST, 
	      paneBuffer_var->instancevars->__o_value,
	      gc_value->__o_value, d_buf);
  } else {
    make_req (shm_mem, PANE_PUT_STR_REQUEST, 
	      win_id_value->__o_value,
	      gc_value->__o_value, d_buf);
  }
#ifdef GRAPHICS_WRITE_SEND_EVENT
  send_event.xgraphicsexpose.type = GraphicsExpose;
  send_event.xgraphicsexpose.send_event = True;
  send_event.xgraphicsexpose.display = display;
  send_event.xgraphicsexpose.drawable = atoi(win_id_value->__o_value);
  XSendEvent (display, atoi(win_id_value->__o_value),
 	      False, 0L, &send_event);
#endif
  wait_req (shm_mem);
#if X11LIB_PRINT_COPIOUS_DEBUG_INFO    
  fprintf (stderr, "__ctalkX11LibPutStr: shm_mem = %s, %s, %s.\n", shm_mem, 
	   &shm_mem[SHM_GC], &shm_mem[SHM_DATA]);
#endif
  return SUCCESS;
}
#endif /* X11LIB_FRAME */
#else /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
static void __gui_support_error (void) {
  _error ("__ctalkGUI ... () functions require Graphical User Interface support.\n");
}
__ctalkX11PanePutStr (OBJECT *self_object, int x, int y, char *s) {
  return __ctalkGUIPaneClearWindow (self_object); 
}
int __ctalkGUIPanePutStr (OBJECT *self_object, int x, int y, char *s) {
  __gui_support_error(); return ERROR;
}
#endif /* ! defined (DJGPP) && ! defined (WITHOUT_X11) */
