/* $Id: rt_obj.c,v 1.4 2010/02/19 11:44:37 kiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2010  Robert Kiesling, ctalk@ctalklang.org.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "rtinfo.h"

extern VARENTRY *__ctalk_dictionary;
extern VARENTRY *__ctalk_last_object;

extern I_PASS interpreter_pass;             /* Declared in rtinfo.c.   */
extern RT_INFO *__call_stack[MAXARGS+1];
extern int __call_stack_ptr;

extern int object_become_delete;
extern int object_become_copy;

/* 
 *   TO DO - For now, if the class name is NULL, simply return 
 *   the first object that matches the name.  Add a 
 *   contextualization facility that allows the language to 
 *   qualify an ambiguous object with its class name.
 *
 *   NOTE - The return object should be checked in cases where,
 *   "name," refers to a method parameter name.
 *
 *   FIXME! -
 *     1. If, "name," refers to a parameter, also check local
 *        objects if a parameter is matched.
 */
OBJECT *__ctalk_get_object (const char *__name, const char *__classname) {

  OBJECT *__o,
    *__rcvr_obj,
    *(*parent_method_fn)();
  VARENTRY *__v;
  METHOD *__m;
  RT_FN *rt_fn;
  int i;

  /*
   *  Check for "self" keyword and parameter replacements.
   */

  if (!strcmp (__name, "self")) {
    __o = __ctalk_self_internal ();
    if (__classname) {
      if (!strcmp (__o -> __o_classname, __classname))
	return __o;
      else
	return NULL;
    } else {
      return __o;
    }
  } else {
    if (!strncmp (__name, "__ctalk_arg_internal", 
		  strlen ("__ctalk_arg_internal"))) {
      int arg_n;
      sscanf (__name, "__ctalk_arg_internal (%d)", &arg_n);
      return __ctalk_arg_internal (arg_n);
    }
  }

  /* 
   * Look for local object in a method's argument block. 
   */
  if (__ctalkIsArgBlockScope ()) {
    if ((__m = __ctalkBlockCallerMethod ()) != NULL) {
      for (i = 0; i < __m -> n_params; i++) {
	if (!strcmp (__m -> params[i] -> name, __name))
	  return __m -> args[__m -> n_params - i - 1] -> obj;
      }
      if (__m -> local_objects.vars) {
	for (__v = __m -> local_objects.vars; __v; __v = __v -> next) {
	  if (__classname) {
	    if (!strcmp (__v -> var_decl -> name, __name) && 
		!strcmp (__v -> var_decl -> class, __classname))
	      return __v -> var_object;
	  } else {
	    if (!strcmp (__v -> var_decl -> name, __name))
	      return __v -> var_object;
	  }
	}
      }
    }
  }

  /*
   *  If we're in a method, try to return a parameter, then local
   *  object. 
   */
  if (((__rcvr_obj = __ctalkRtMethodClassObject ()) != NULL) &&
      ((parent_method_fn = __ctalkRtGetMethodFn ()) != NULL)) {
    if ((__m = __ctalkRtGetMethod ()) != NULL) {
      for (i = 0; i < __m -> n_params; i++) {
	if (!strcmp (__m -> params[i] -> name, __name))
	  return __m -> args[__m -> n_params - i - 1] -> obj;
      }
      if (__m -> local_objects.vars) {
	for (__v = __m -> local_objects.vars; __v; __v = __v -> next) {
	  if (__classname) {
	    if (!strcmp (__v -> var_decl -> name, __name) && 
		!strcmp (__v -> var_decl -> class, __classname))
	      return __v -> var_object;
	  } else {
	    if (!strcmp (__v -> var_decl -> name, __name))
	      return __v -> var_object;
	  }
	}
      }
    }
  }

  /*
   *  If not found, look up the parent method on the call stack if 
   *  the object is local to an inline method caller like map.
   */
  if (__call_stack_ptr < MAXARGS) {
    if (((__rcvr_obj = __call_stack[__call_stack_ptr+1]->rcvr_obj) != NULL) &&
	((parent_method_fn = __call_stack[__call_stack_ptr+1]->method_fn) 
	 != NULL)) {
      if ((__m = __ctalkRtGetMethod ()) != NULL) {
	int __arg_stack_ptr, __arg_frame_idx, __nth_arg_frame_offset,
	  __nth_arg;
	OBJECT *__arg_object;
	__arg_stack_ptr = __ctalkGetArgPtr ();
	/*
	 *  Handle args for methods called recursively by 
	 *  checking each method arg against the argument 
	 *  stack's. current frame.  Note that the 
	 *  method's arg_frame_top may not always be the
	 *  same relatively (sometimes it might point to the
	 *  next arg index _after_ the call), so we don't use 
	 *  it as a  bounds when checking the actual arguments 
	 *  on the __arg_stack.
	 *
	 *  If there's not arg_frame_top given, then simply
	 *  check the first args[0... n_params].
	 */
	if (__m -> n_params && 
	    (__m -> arg_frame_top >= __arg_stack_ptr) &&
	    (__m -> arg_frame_top <= MAXARGS) &&
	    (!__m -> varargs)) {
	  __nth_arg_frame_offset = 
	    (__m -> n_args == __m -> n_params) ?
	    0 : ((__m -> n_args / __m -> n_params) - 1);
	next_arg_frame:
	  for (i = 0; i < __m -> n_params; i++) {
	    if ((__m -> params[i] != NULL) && 
		!strcmp (__m -> params[i] -> name, __name)) {
	      __nth_arg = 
		(__m->n_params-i-1)+__nth_arg_frame_offset;
	      if (__nth_arg < 0)
		goto look_at_vraiables;
	      __arg_object = 
		(((__nth_arg >= 0) && __m->args[__nth_arg]) ?
		 __m -> args[__nth_arg]->obj : NULL);
	      for (__arg_frame_idx = __arg_stack_ptr + 1; 
		   __arg_frame_idx <= MAXARGS;
		   __arg_frame_idx++) {
		if (__arg_object == __ctalkArgAt (__arg_frame_idx)) {
		  return __arg_object;
		}
	      }
	      __nth_arg_frame_offset -= __m -> n_params;
	      goto next_arg_frame;
	    }
	  }
	} else {
	  for (i = 0; i < __m -> n_params; i++) {
	    if ((__m -> params[i] != NULL) && 
		!strcmp (__m -> params[i] -> name, __name))
	      return __m -> args[__m -> n_params - i - 1] -> obj;
	  }
	}

      look_at_vraiables:
	if (IS_CONSTRUCTOR(__m) &&
	    !strcmp (__m->params[0]->name, __name)) {
	  if (__m->local_objects.vars)
	    return __m->local_objects.vars->var_object;
	}
	if (__m -> local_objects.vars) {
	  for (__v = __m -> local_objects.vars; 
	       __v && __v -> var_object && IS_OBJECT(__v->var_object); 
	       __v = __v -> next) {
	    if (__classname) {
	      if (!strcmp (__v -> var_decl -> name, __name) && 
		  !strcmp (__v -> var_decl -> class, __classname))
		return __v -> var_object;
	    } else {
	      if (!strcmp (__v -> var_decl -> name, __name))
		return __v -> var_object;
	    }
	  }
	}
      }

      /*
       *  Check for a global object.
       */

      if (__ctalk_dictionary) {
	for (__v = __ctalk_dictionary; __v; __v = __v -> next) {
	  if (__classname) {
	    if (!strcmp (__v -> var_decl -> name, __name) &&
		!strcmp (__v -> var_decl -> class, __classname))
	      return __v -> var_object;
	  } else {
	    if (!strcmp (__v -> var_decl -> name, __name))
	      return __v -> var_object;
	  }
	  if (!__v || !__v -> next)
	    break;
	}
      }

      /*
       *  Look for an object declared in a function and used
       *  within an argument block.
       */
      if (__ctalkIsArgBlockScope ()) {
	if ((rt_fn = __ctalkBlockCallerFn ()) != NULL) {
	  for (__v = rt_fn -> local_objects.vars; __v; __v = __v -> next) {
	    if (__classname) {
	      if (!strcmp (__name, __v -> var_decl -> name) && 
		  !strcmp (__classname, __v -> var_decl -> class))
		return __v -> var_object;
	    } else {
	      if (!strcmp (__name, __v -> var_decl -> name))
		return __v -> var_object;
	    }
	  }
	}
      }

      if ((rt_fn = get_fn ()) != NULL) {
	for (__v = rt_fn -> local_objects.vars; __v; __v = __v -> next) {
	  if (__classname) {
	    if (!strcmp (__name, __v -> var_decl -> name) && 
		!strcmp (__classname, __v -> var_decl -> class))
	      return __v -> var_object;
	  } else {
	    if (!strcmp (__name, __v -> var_decl -> name))
	      return __v -> var_object;
	  }
	}
      }

      /*
       * Also check for a class object.  TO DO - This is a duplication
       * of the code below, and should be factored out 
       * if this function gets more complicated.
       */

      if (__classname && !strcmp (__classname, "Class")) {
	return __ctalkGetClass (__name);
      }
      return NULL;
    }
  }

  /*
   *  If we're in a function, try to find the object there.
   */
  if (interpreter_pass == run_time_pass) {
    if ((rt_fn = get_fn ()) != NULL) {
      for (__v = rt_fn -> local_objects.vars; __v; __v = __v -> next) {
 	if (__classname) {
 	  if (!strcmp (__name, __v -> var_decl -> name) && 
 	      !strcmp (__classname, __v -> var_decl -> class))
 	    return __v -> var_object;
 	} else {
 	  if (!strcmp (__name, __v -> var_decl -> name))
 	    return __v -> var_object;
 	}
      }
    }
  }

  if (__classname && !strcmp (__classname, "Class")) {
    return __ctalkGetClass (__name);
  } else {
    for (__v = __ctalk_dictionary; __v; __v = __v -> next) {
      if (__classname) {
	if (!strcmp (__v -> var_decl -> name, __name) &&
	    !strcmp (__v -> var_decl -> class, __classname))
	  return __v -> var_object;
      } else {
	if (!strcmp (__v -> var_decl -> name, __name))
	  return __v -> var_object;
      }
    }
  }

  /*
   *  Check for a class on its own.  Only CFunction class 
   *  methods need this at the moment.
   */
  if ((__o = __ctalkGetClass (__name)) != NULL)
    return __o;

  /*
   *  Finally, check for a class variable.
   */
  if ((__o = __ctalk_find_classvar (__name, __classname)) != NULL)
    return __o;

  return NULL;
}

/*
 *  This function should not need to check for, "self," and, "super,"
 *  because at this time it is only called from 
 *  __ctalkDeleteObjectInternal () and __ctalk_set_local ().
 */
int __ctalk_remove_object (OBJECT *o) {

  if (IS_CLASS_OBJECT(o)) {
    __ctalkRemoveClass (o);
  } else {
    if (__ctalk_last_object && (o == __ctalk_last_object -> var_object)) {
      if (object_become_delete) {
	__ctalk_last_object -> var_object = NULL;
      } else {
	__ctalk_last_object = __ctalk_last_object -> prev;
      }
    }
    if (__ctalk_dictionary && (o == __ctalk_dictionary -> var_object)) {
      if (object_become_delete) {
	__ctalk_dictionary -> var_object = NULL;
      } else {
	__ctalk_dictionary = __ctalk_dictionary -> next;
      }
    }
    if (o -> prev) o -> prev -> next = o -> next;
    if (o -> next) o -> next -> prev = o -> prev;
    o -> next = o -> prev = NULL;
  }

  return SUCCESS;
}

OBJECT *__ctalk_self_internal (void) {

  OBJECT *self_obj;

  self_obj = __ctalk_receiver_pop ();
  __ctalk_receiver_push (self_obj);
  return self_obj;
}

void __ctalk_set_global (char *name, char *classname) {

  OBJECT *obj;

  if ((obj = __ctalk_get_object (name, classname)) == NULL)
    _error ("__ctalk_set_global (): Undefined object %s (Class %s).\n", 
	    name, ((classname) ? classname : NULLSTR));

  __objRefCntSet (OBJREF(obj), 1);
  __ctalkSetObjectScope (obj, GLOBAL_VAR);

}


#define __THIS_METHOD (__call_stack[__call_stack_ptr+1]->method)
void __ctalk_set_local (OBJECT *obj) {

  RT_FN *rt_fn;
  VARENTRY *v, *v_ptr;

  if (!obj || !IS_OBJECT(obj))
    return;

  __ctalkSetObjectScope (obj, LOCAL_VAR);

  __objRefCntSet(OBJREF(obj), 1);

  /*
   *  If we're in a method, retrieve the method, and add the object
   *  there.
   */
  if (__THIS_METHOD) {
    __ctalk_remove_object (obj);
    obj -> next = obj -> prev = NULL;
    if (obj -> __o_varentry) {
      if (obj -> __o_varentry -> next)
	obj -> __o_varentry -> next -> prev = obj -> __o_varentry -> prev;
      if (obj -> __o_varentry -> prev)
	obj -> __o_varentry -> prev -> next = obj -> __o_varentry -> next;
      delete_varentry (obj -> __o_varentry);
      obj -> __o_varentry = NULL;
    }

    v = new_varentry (obj);
    if (__THIS_METHOD -> local_objects.vars) {
      for ( v_ptr = __THIS_METHOD -> local_objects.vars; 
	    v_ptr -> next; v_ptr = v_ptr -> next)
	;
      v_ptr -> next = v;
      v -> prev = v_ptr;
    } else {
      __THIS_METHOD -> local_objects.vars = v;
    }
    return;
  }
  
  /*
   *  If there's no receiver object, then we're in a C function.
   */

  if ((rt_fn = get_fn ()) == NULL) 
    return;

  __ctalk_remove_object (obj);

  if (obj -> __o_varentry) {
    if (obj -> __o_varentry -> next)
      obj -> __o_varentry -> next -> prev = obj -> __o_varentry -> prev;
    if (obj -> __o_varentry -> prev)
      obj -> __o_varentry -> prev -> next = obj -> __o_varentry -> next;
    delete_varentry (obj -> __o_varentry);
    obj -> __o_varentry = NULL;
  }

  obj -> next = obj -> prev = NULL;

  if (rt_fn -> local_objects.vars) {
    VARENTRY *v_1;
    v = new_varentry (obj);

    for (v_1 = rt_fn -> local_objects.vars; v_1 -> next; v_1 = v_1 -> next)
      ;
    v_1 -> next = v;
    v -> prev = v_1;

  } else {
    v = new_varentry (obj);
    rt_fn -> local_objects.vars = v;
  }
}

OBJECT *match_instancevar (OBJECT *varlist, OBJECT *var) {
  OBJECT *o, *p;
  for (o = varlist; o; o = o -> next) {
    if (o == var) {
      return o;
    } else {
      if (o -> instancevars)
	if ((p = match_instancevar (o -> instancevars, var)) != NULL)
	  return p;
    }
  }

  return NULL;
}

/*
 *  TO DO - Only used by __ctalkDeleteObjectInternal.  Should be 
 *  generalized.
 */
OBJECT *__ctalkInstanceVariableObject (OBJECT *var) {

  OBJECT *r;
  VARENTRY *v;

  for (v = __ctalk_dictionary; v; v = v -> next) {
    if ((r = match_instancevar (v -> var_object -> instancevars, var)) != NULL)
      return v -> var_object;
  }
  return NULL;
}

/*
 *  Like __ctalkInstanceVariableObject, but checks for 
 *  instance variable membership of objects on the receiver
 *  stack.
 */

OBJECT *__ctalkInstanceVariableReceiver (OBJECT *possible_var) {

  int i;
  OBJECT *rcvr, *v;

  for (i = __ctalkGetReceiverPtr () + 1; i <= MAXARGS; i++) {
    rcvr = __ctalk_receivers_at (i);
    for (v = rcvr -> instancevars; v; v = v -> next) {
      if (v == possible_var)
	return rcvr;
    }
  }
  return possible_var;
}

int copy_object_lvl = 0;
int max_copy_lvl_reached = 0;

/*
 *  This is the original version - makes no extra provision
 *  for a value and following instance variables to be copied
 *  individually.
 */

static OBJECT *__copy_object_internal (OBJREF_T, OBJREF_T);

static OBJECT *__copy_object_internal_from_parent (OBJREF_T src_ptr, OBJREF_T dest_ptr) {
  OBJECT *p_var, *p_var2, *new_var, *reffed_obj;

  if (copy_object_lvl >= MAXARGS) {
    fprintf (stderr, "__ctalkCopyObject: %s: Maximum object nesting level reached.\n", (*src_ptr)->__o_name);
    __warning_trace ();
    max_copy_lvl_reached = TRUE;
    --copy_object_lvl;
    return NULL;
  }

  if (max_copy_lvl_reached)
    return NULL;

  ++copy_object_lvl;
  if (*src_ptr == NULL)  {
    _warning ("__ctalk_copy_object_internal: NULL source object.\n");
    return NULL;
  }
  *dest_ptr = __ctalkCreateObject ((*src_ptr) -> __o_name, 
				   (*src_ptr) -> __o_classname, 
				   (((*src_ptr) -> __o_superclassname) ? 
				    (*src_ptr) -> __o_superclassname : NULL),
				   (*src_ptr) -> scope);
  if (((reffed_obj = obj_ref_str((*src_ptr)->__o_value)) != NULL) &&
      !object_become_copy) {
    if (!parent_ref_is_circular (*src_ptr, reffed_obj)) {
      __ctalkSetObjectValue 
	(*dest_ptr, __ctalkCopyObjectRef
	 (*src_ptr, (*src_ptr)->__o_value));
    } else {
      /*
       *  It's not always possible to do a fixup here, due to the 
       *  fact that the destination object may not be fixed up
       *  until much later, and it's not possible to determine
       *  which of the destination objects the parent objects 
       *  reference.  So, just issue a warning here.
       */
      _warning ("__copy_object_internal: Circular object reference in %s.\n",
 		((*src_ptr)->__o_p_obj ? (*src_ptr)->__o_p_obj->__o_name :
 		 (*src_ptr)->__o_name));
      __ctalkSetObjectValue (*dest_ptr, (*src_ptr) -> __o_value);
    }
  } else {
    __ctalkSetObjectValue (*dest_ptr, (*src_ptr) -> __o_value);
  }

  for (p_var = (*src_ptr) -> instancevars; p_var; p_var = p_var -> next) {
    /*
     *  Don't use __ctalkAddInstanceVariable here because some
     *  classes (e.g. List) have duplicate instance variable names.
     *
     *  Can't use IS_VALUE_INSTANCE_VAR macro when __o_p_obj is
     *  not always set.  This function is also used by the front
     *  end.
     */
    if (__copy_object_internal_from_parent (OBJREF(p_var), OBJREF(new_var))) {
      new_var -> __o_p_obj = (*dest_ptr);
      if ((*dest_ptr) -> instancevars && 
	  IS_OBJECT((*dest_ptr)->instancevars)) {
	for (p_var2 = (*dest_ptr)->instancevars; p_var2 && p_var2 -> next;
	     p_var2 = p_var2 -> next) {
	  if (!p_var2 -> next)
	    break;
	}
	p_var2 -> next = new_var;
	new_var -> prev = p_var2;
	__objRefCntSet (OBJREF(new_var), p_var -> nrefs);
      } else {
	(*dest_ptr)->instancevars = new_var;
	__objRefCntSet (OBJREF(new_var), (*src_ptr) -> nrefs);
      }
    } else {
      --copy_object_lvl;
      return NULL;
    }
  }
  --copy_object_lvl;
  return *dest_ptr;
}

static OBJECT *__copy_object_internal (OBJREF_T src_ptr, OBJREF_T dest_ptr) {
  OBJECT *p_var, *p_var2, *new_var, *reffed_obj;

  if (copy_object_lvl >= MAXARGS) {
    fprintf (stderr, "__ctalkCopyObject: %s: Maximum object nesting level reached.\n", (*src_ptr)->__o_name);
    __warning_trace ();
    max_copy_lvl_reached = TRUE;
    --copy_object_lvl;
    return NULL;
  }

  if (max_copy_lvl_reached)
    return NULL;

  ++copy_object_lvl;
  if (*src_ptr == NULL)  {
    _warning ("__ctalk_copy_object_internal: NULL source object.\n");
    return NULL;
  }
  *dest_ptr = __ctalkCreateObject ((*src_ptr) -> __o_name, 
				   (*src_ptr) -> __o_classname, 
				   (((*src_ptr) -> __o_superclassname) ? 
				    (*src_ptr) -> __o_superclassname : NULL),
				   (*src_ptr) -> scope);
  if (((reffed_obj = obj_ref_str((*src_ptr)->__o_value)) != NULL) &&
      !object_become_copy) {
    if (!parent_ref_is_circular (*src_ptr, reffed_obj)) {
      __ctalkSetObjectValue 
	(*dest_ptr, __ctalkCopyObjectRef
	 (*src_ptr, (*src_ptr)->__o_value));
    } else {
      /*
       *  It's not always possible to do a fixup here, due to the 
       *  fact that the destination object may not be fixed up
       *  until much later, and it's not possible to determine
       *  which of the destination objects the parent objects 
       *  reference.  So, just issue a warning here.
       */
      _warning ("__copy_object_internal: Circular object reference in %s.\n",
 		((*src_ptr)->__o_p_obj ? (*src_ptr)->__o_p_obj->__o_name :
 		 (*src_ptr)->__o_name));
      __ctalkSetObjectValue (*dest_ptr, (*src_ptr) -> __o_value);
    }
  } else {
    __ctalkSetObjectValue (*dest_ptr, (*src_ptr) -> __o_value);
  }

  if (!IS_VALUE_INSTANCE_VAR (*src_ptr)) {
    for (p_var = (*src_ptr) -> instancevars; p_var; p_var = p_var -> next) {
      /*
       *  Don't use __ctalkAddInstanceVariable here because some
       *  classes (e.g. List) have duplicate instance variable names.
       *
       *  Can't use IS_VALUE_INSTANCE_VAR macro when __o_p_obj is
       *  not always set.  This function is also used by the front
       *  end.
       */
      if (__copy_object_internal_from_parent 
	  (OBJREF(p_var), OBJREF(new_var))) {
	new_var -> __o_p_obj = (*dest_ptr);
	if ((*dest_ptr) -> instancevars && 
	    IS_OBJECT((*dest_ptr)->instancevars)) {
	  for (p_var2 = (*dest_ptr)->instancevars; p_var2 && p_var2 -> next;
	       p_var2 = p_var2 -> next) {
	    if (!p_var2 -> next)
	      break;
	  }
	  p_var2 -> next = new_var;
	  new_var -> prev = p_var2;
	  __objRefCntSet (OBJREF(new_var), p_var -> nrefs);
	} else {
	  (*dest_ptr)->instancevars = new_var;
	  __objRefCntSet (OBJREF(new_var), (*src_ptr) -> nrefs);
	}
      } else {
	--copy_object_lvl;
	return NULL;
      }
    }
  } else { /* if (!IS_VALUE_INSTANCE_VAR (*src_ptr)) { */
    p_var2 = *dest_ptr;
    for (p_var = (*src_ptr) -> next; p_var; p_var = p_var -> next) {
      if (__copy_object_internal (OBJREF(p_var), OBJREF(new_var))) {
	new_var -> __o_p_obj = (*dest_ptr) -> __o_p_obj;
	p_var2 -> next = new_var;
	p_var2 = new_var;
      }
    }
  }
  --copy_object_lvl;
  return *dest_ptr;
}

/*
 *  Return True if reffed contains o.  Look in objects named by
 *  reference strings here also.
 */

int reffed_ref_lvl = 0;
int reffed_ref_lvl_reached = 0;

static int reffed_ref_is_circular_a (OBJECT *o, OBJECT *reffed) {
  OBJECT *t, *value_reffed_obj;
  int r;
  if (reffed_ref_lvl_reached) {
    --reffed_ref_lvl;
    return TRUE;
  }
  ++reffed_ref_lvl;
  if (reffed_ref_lvl >= MAXARGS) {
    reffed_ref_lvl_reached = TRUE;
    return TRUE;
  }
  for (t = reffed -> instancevars; t; t = t -> next) {
    if (t -> instancevars) {
      if ((r = reffed_ref_is_circular_a (o, t)) != FALSE) {
	--reffed_ref_lvl;
	return r;
      }
    }
    if ((value_reffed_obj = obj_ref_str (t -> __o_value)) != NULL) {
      if ((r = reffed_ref_is_circular_a (o, value_reffed_obj)) != FALSE) {
	--reffed_ref_lvl;
	return r;
      }
    }
    if (t == o) {
      --reffed_ref_lvl;
      return TRUE;
    }
  }
  --reffed_ref_lvl;
  return FALSE;
}

static int reffed_ref_is_circular (OBJECT *o, OBJECT *reffed) {
  int r;
  reffed_ref_lvl = reffed_ref_lvl_reached = 0;
  r = reffed_ref_is_circular_a (o, reffed);
  reffed_ref_lvl = reffed_ref_lvl_reached = 0;
  return r;
}

char *__ctalkCopyObjectRef(OBJECT *src_obj, char *__obj_ref_str) {
  OBJECT *src, *dest;
  static char dest_buf[MAXLABEL];
  sscanf (__obj_ref_str, "0x%p", &src);
  if (reffed_ref_is_circular (src_obj, src)) {
    __objRefCntInc (OBJREF(src));
    return __obj_ref_str;
  }
  if (!src || !IS_OBJECT(src)) {
    dest = 
      __ctalkCreateObjectInit (NULLSTR, 
			       INTEGER_CLASSNAME,
			       INTEGER_SUPERCLASSNAME, 
			       LOCAL_VAR, "0");
    sprintf (dest_buf, "%#x", (unsigned int)dest);
    return dest_buf;
  }
  /*
   *  Class objects often refer to themselves.
   */
  if (IS_CLASS_OBJECT(src))
    return __obj_ref_str;
  if (src -> __o_value) {
    if (!strcmp (__obj_ref_str, src->__o_value)) {
      if (src -> instancevars) {
	if (!strcmp (__obj_ref_str, src->instancevars->__o_value)) {
	  strcpy (dest_buf, __obj_ref_str);
	  return dest_buf;
	}
      } else {
	strcpy (dest_buf, __obj_ref_str);
	return dest_buf;
      }
    }
  }
  if (__copy_object_internal (OBJREF(src), OBJREF(dest))) {
    __objRefCntSet(OBJREF(dest), src->nrefs);
    sprintf (dest_buf, "%#x", (unsigned int)dest);
    return dest_buf;
  } else {
    return NULL;
  }
}

char *__ctalkCopyObjectRef_old(char *__obj_ref_str) {
  OBJECT *src, *dest;
  static char dest_buf[MAXLABEL];
  sscanf (__obj_ref_str, "0x%p", &src);
  if (!src || !IS_OBJECT(src)) {
    dest = 
      __ctalkCreateObjectInit (NULLSTR, 
			       INTEGER_CLASSNAME,
			       INTEGER_SUPERCLASSNAME, 
			       LOCAL_VAR, "0");
    sprintf (dest_buf, "%#x", (unsigned int)dest);
    return dest_buf;
  }
  /*
   *  Class objects often refer to themselves.
   */
  if (IS_CLASS_OBJECT(src))
    return __obj_ref_str;
  if (src -> __o_value) {
    if (!strcmp (__obj_ref_str, src->__o_value)) {
      if (src -> instancevars) {
	if (!strcmp (__obj_ref_str, src->instancevars->__o_value)) {
	  strcpy (dest_buf, __obj_ref_str);
	  return dest_buf;
	}
      } else {
	strcpy (dest_buf, __obj_ref_str);
	return dest_buf;
      }
    }
  }
  if (__copy_object_internal (OBJREF(src), OBJREF(dest))) {
    __objRefCntSet(OBJREF(dest), src->nrefs);
    sprintf (dest_buf, "%#x", (unsigned int)dest);
    return dest_buf;
  } else {
    return NULL;
  }
}

OBJECT *__ctalkCopyObject (OBJREF_T src_ptr, OBJREF_T dest_ptr) {
  OBJECT *result;
  copy_object_lvl = max_copy_lvl_reached = 0;
  result = __copy_object_internal (src_ptr, dest_ptr);
  copy_object_lvl = max_copy_lvl_reached = 0;
  return result;
}

void __ctalkSetObjectScope (OBJECT *__o, int scope) {

  OBJECT *__var;
  if (__o && IS_OBJECT(__o)) {
    for (__var = __o -> instancevars; __var; __var = __var -> next) 
      __ctalkSetObjectScope (__var, scope);
    __o -> scope = scope;
  }
}

extern int __local_object_init;

int object_is_deletable (OBJECT *obj, OBJECT *result_obj, 
			 OBJECT *subexpr_obj){
  if ((obj -> scope & LOCAL_VAR) &&
      (__local_object_init == TRUE)) {
    if ((obj != result_obj) &&
	(obj != subexpr_obj)) {
      return TRUE;
    }
    if (obj -> scope & METHOD_USER_OBJECT)
      return FALSE;
  }
  if ((obj -> scope == CREATED_PARAM) &&
      (obj != result_obj) &&
      (obj != subexpr_obj) &&
      (obj -> instancevars != result_obj)) {
    return TRUE;
  }
  if ((obj -> scope & CREATED_PARAM) && (obj -> nrefs == 0) && 
      (obj != result_obj) &&
      (obj != subexpr_obj) &&
      (obj -> instancevars != result_obj)) {
    return TRUE;
  }
  /* Strange but true... */
  if ((obj -> scope & CVAR_VAR_ALIAS_COPY) &&
      !(obj -> scope & LOCAL_VAR) &&
      (obj != result_obj) &&
      (obj != subexpr_obj))
    return TRUE;
  return FALSE;
}

int object_is_decrementable (OBJECT *obj, OBJECT *result_obj,
			     OBJECT *subexpr_obj) {
  if ((obj -> scope & CVAR_VAR_ALIAS_COPY)&&
      (obj != result_obj) &&
      (obj != subexpr_obj))
    return TRUE;
  if ((obj -> __o_p_obj) && 
      (obj -> nrefs <= obj -> __o_p_obj -> nrefs))
    return FALSE;
  if (is_p_obj (obj, result_obj) ||
      is_p_obj (obj, subexpr_obj))
    return FALSE;
  if ((obj -> scope & METHOD_USER_OBJECT) &&
      (obj -> nrefs <= 1))
    return FALSE;
  if ((obj -> scope & TYPECAST_OBJECT) &&
      (obj != result_obj) &&
      (obj != subexpr_obj))
    return TRUE;
  if ((obj -> scope & ARG_VAR) &&
      ((obj == result_obj) || (obj == subexpr_obj)))
    return FALSE;
  if (!obj -> __o_varentry) {
    if (obj && IS_OBJECT (obj) && 
	(obj != result_obj) && 
	(obj != subexpr_obj) && 
	!__ctalk_get_object(obj->__o_name, obj->__o_classname) &&
	!__ctalkFindClassVariable (obj->__o_name, FALSE) &&
	!is_receiver (obj) &&
	!is_arg (obj)) {
      return TRUE;
    }
  } else {
    if (obj && IS_OBJECT (obj) && 
	(obj != result_obj) && 
	(obj != subexpr_obj) && 
	!__ctalk_get_object(obj->__o_varentry->var_decl->name, 
			    obj->__o_varentry->var_decl->class) &&
	!__ctalkFindClassVariable (obj->__o_varentry->var_decl->name, 
				   FALSE) &&
	!is_receiver (obj) &&
	!is_arg (obj)) {
      return TRUE;
    }
  }
  if (IS_CLASS_OBJECT(obj)) {
    OBJECT *__o_class;
    if ((__o_class = __ctalkGetClass (obj -> __o_name)) != NULL) {
      if (__o_class != obj) {
	if (obj -> scope & CVAR_VAR_ALIAS_COPY) {
	  return FALSE;
	} else {
	  if (obj -> scope & ARG_VAR) {
	    if ((obj == result_obj) || (obj == subexpr_obj)) {
	      return FALSE;
	    } else {
	      return TRUE;
	    }
	  } else{
	    return TRUE;
	  }
	}
      }
    }
  }
  return FALSE;
}

VARENTRY *new_varentry (OBJECT *var_object) {
  VARENTRY *v;
  PARAM *p;
  v = (VARENTRY *)__xalloc (sizeof (struct _varentry));
  p = new_param ();
  strcpy (p -> class, var_object -> __o_classname);
  strcpy (p -> name, var_object -> __o_name);
  v -> var_decl = p;
  v -> var_object = var_object;
  var_object -> __o_varentry = v;
  return v;
}

void delete_varentry (VARENTRY *v) {
  __xfree (v -> var_decl);
  __xfree (v);
}

OBJECT *__ctalkReplaceVarEntry (VARENTRY *v, OBJECT *new_object) {
  OBJECT *old_object;
  old_object = v -> var_object;
  old_object -> __o_varentry = NULL;
  v -> var_object = new_object;
  new_object -> __o_varentry = v;
  return old_object;
}

void __delete_operand_result (OBJREF_T operand_result_p, 
			      OBJREF_T subexpr_result_p) {
  if ((*operand_result_p) &&
      IS_OBJECT((*operand_result_p)) &&
      ((*operand_result_p) -> scope & SUBEXPR_CREATED_RESULT)) {
    (*operand_result_p) -> scope &= ~SUBEXPR_CREATED_RESULT;
    __objRefCntDec (operand_result_p);
    if ((*operand_result_p) -> nrefs <= 0) {
      __ctalkDeleteObjectInternal ((*operand_result_p));
      if ((*operand_result_p) == (*subexpr_result_p)) {
	(*operand_result_p) = (*subexpr_result_p) = NULL;
      } else {
	(*operand_result_p) = NULL;
      }
    } else {
      __ctalkDeleteObjectInternal ((*operand_result_p));
    }
  }
}

static void delete_reference_in_user_object (OBJECT *user_object) {
  OBJECT *r;
  if ((r = obj_ref_str ((user_object -> instancevars) ?
		   user_object -> instancevars -> __o_value :
			user_object -> __o_value)) != NULL) {
    if (user_object -> nrefs == 0) {
      /*
       *  Don't remove the reference if it's to a temporary
       *  object also, then delete the referenced object itself
       *  further on.
       */
      if ((r -> __o_varentry == 0) &&
	  (r -> nrefs <= 0))
	return;
      if (user_object -> instancevars) {
	user_object -> instancevars -> __o_value[0] = '\0';
      } else {
	user_object -> __o_value[0] = '\0';
      }
    }
  }
  return;
}

int __ctalkRegisterUserObject (OBJECT *o) {
  METHOD *m;
  OBJECT *o_1, *o_top;
  int n_resource_objs;

  if (!o || !IS_OBJECT(o))
    return SUCCESS;

  if (IS_CLASS_OBJECT(o) && __ctalkGetClass (o -> __o_name))
    return SUCCESS;

  if (o -> scope & GLOBAL_VAR)
    return SUCCESS;

  if (o -> scope & METHOD_USER_OBJECT) {
    __objRefCntInc (OBJREF(o));
    return SUCCESS;
  }

  if (((o_top = top_parent_object (o)) != NULL) &&
      ((o_top -> __o_varentry) || IS_CLASS_OBJECT(o_top)))
    return SUCCESS;

  if ((m = __ctalkRtGetMethod ()) != NULL) {
    __objRefCntSet (OBJREF(o), 1);
    __ctalkSetObjectScope (o, o -> scope | METHOD_USER_OBJECT);
    if (m -> user_objects == NULL) {
      m -> user_objects = m -> user_object_ptr = o;
    } else {
      for (o_1 = m -> user_objects, 
	     n_resource_objs = 1;
	   o_1 && IS_OBJECT(o_1);
	   o_1 = o_1 -> next, n_resource_objs++) 
	;
      if (n_resource_objs >= MAX_USER_OBJECT_RESOURCES) {
	OBJECT *o_del;
	o_del = m -> user_objects;
	m -> user_objects = o_del -> next;
	m -> user_objects -> prev = 
	  o_del -> next = NULL;
	__objRefCntDec(OBJREF(o_del));
	delete_reference_in_user_object (o_del);
	__ctalkDeleteObjectInternal (o_del);
      }
      m -> user_object_ptr -> next = o;
      o -> prev = m -> user_object_ptr;
      m -> user_object_ptr = o;
      return SUCCESS;
    }
  }
  return SUCCESS;
}

int __ctalkRegisterExtraObject (OBJECT *o) {
  METHOD *m;
  OBJECT *o_1, *o_top;
  int n_resource_objs;

  if (IS_CLASS_OBJECT(o) && __ctalkGetClass (o->__o_name))
    return SUCCESS;

  if (o -> scope & GLOBAL_VAR)
    return SUCCESS;

  if (o -> scope & METHOD_USER_OBJECT) {
    __objRefCntInc (OBJREF(o));
    return SUCCESS;
  }

  if (((o_top = top_parent_object (o)) != NULL) &&
      ((o_top -> __o_varentry) || IS_CLASS_OBJECT(o_top)))
    return SUCCESS;

  if ((m = __ctalkRtGetMethod ()) != NULL) {
    if (m -> user_objects == NULL) {
      m -> user_objects = m -> user_object_ptr = o;
    } else {
      for (o_1 = m -> user_objects, 
	     n_resource_objs = 1;
	   o_1;
	   o_1 = o_1 -> next, n_resource_objs++)  {
	if (o == o_1)
	  return SUCCESS;
      }
      if (n_resource_objs >= MAX_USER_OBJECT_RESOURCES) {
	OBJECT *o_del;
	o_del = m -> user_objects;
	m -> user_objects = o_del -> next;
	m -> user_objects -> prev = NULL;
	__objRefCntDec(OBJREF(o_del));
	delete_reference_in_user_object (o_del);
	__ctalkDeleteObjectInternal (o_del);
      }
      m -> user_object_ptr -> next = o;
      o -> prev = m -> user_object_ptr;
      m -> user_object_ptr = o;
      return SUCCESS;
    }
  }
  return SUCCESS;
}

int is_p_obj (OBJECT *p_obj, OBJECT *c_obj) {
  OBJECT *o;
  if (!c_obj || 
      !IS_OBJECT(c_obj) ||
      (c_obj -> __o_p_obj == NULL) ||
      !IS_OBJECT(c_obj -> __o_p_obj))
    return FALSE;
  for (o = c_obj -> __o_p_obj; o && IS_OBJECT(o); o = o -> __o_p_obj) {
    if (o == p_obj)
      return TRUE;
  }
  return FALSE;
}

int parent_ref_is_circular (OBJECT *o, OBJECT *ref) {
  OBJECT *t;
  for (t = o -> __o_p_obj; t; t = t -> __o_p_obj)
    if (t == ref)
      return TRUE;
  return FALSE;
}

OBJECT *top_parent_object (OBJECT *o) {
  OBJECT *t;
  if (!o -> __o_p_obj)
    return NULL;
  for (t = o -> __o_p_obj; t && t -> __o_p_obj; t = t -> __o_p_obj)
    ;
  return t;
}
