/* $Id: rt_vmthd.c,v 1.5 2010/03/03 03:10:10 kiesling Exp $ */

/*
  This file is part of Ctalk.
  Copyright  2005-2010  Robert Kiesling, ctalk@ctalklang.org.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "rtinfo.h"
#include "typeof.h"
#include "list.h"

extern RT_INFO *__call_stack[MAXARGS+1];

static int __method_object_init (OBJECT *rcvr_obj, OBJECT *class_obj, 
				 METHOD *actual_method) {
  /*
   *  The instance variables are defined in the class/Method library
   *  file.  
   *  
   *  Note that this function does not yet fill in the parameters
   *  and arguments, nor local_objects, user_objects, local_cvars,
   *  imported, error_line, and error_column members.
   */
  OBJECT *name_instance_var, *selector_instance_var,
    *returnClass_instance_var, *rcvrClassObject_instance_var,
    *methodFn_instance_var, *source_instance_var, 
    *nParams_instance_var, *varargs_instance_var, *nArgs_instance_var,
    *argFrameTop_instance_var, *rcvrFrameTop_instance_var;
  char s[MAXLABEL];
  if ((name_instance_var = 
       __ctalkGetInstanceVariable (rcvr_obj, "methodName", TRUE)) == NULL)
    return ERROR;
  __ctalkSetObjectValueVar (name_instance_var, actual_method -> name);
  if (*(actual_method -> selector)) {
    if ((selector_instance_var = 
	 __ctalkGetInstanceVariable (rcvr_obj, "methodSelector", TRUE)) == NULL)
      return ERROR;
    __ctalkSetObjectValueVar (selector_instance_var, 
			      actual_method->selector);
  }
  if (*(actual_method -> returnclass)) {
    if ((returnClass_instance_var = 
	 __ctalkGetInstanceVariable (rcvr_obj, "returnClass", TRUE)) == NULL)
      return ERROR;
    __ctalkSetObjectValueVar (returnClass_instance_var, 
			      actual_method->returnclass);
  }
  if ((rcvrClassObject_instance_var = 
       __ctalkGetInstanceVariable (rcvr_obj, "rcvrClassObject", 
				   TRUE)) == NULL)
    return ERROR;
  sprintf (s, "%#x", (unsigned int)actual_method->rcvr_class_obj);
  __ctalkSetObjectValueVar (rcvrClassObject_instance_var, s);
  if ((methodFn_instance_var = 
       __ctalkGetInstanceVariable (rcvr_obj, "methodFn", TRUE)) == NULL)
    return ERROR;
  sprintf (s, "%#x", (unsigned int)actual_method->cfunc);
  __ctalkSetObjectValueVar (methodFn_instance_var, s);
  if (actual_method -> src != NULL) {
    if ((source_instance_var = 
	 __ctalkGetInstanceVariable (rcvr_obj, "methodSource", TRUE)) == NULL)
      return ERROR;
    __ctalkSetObjectValueVar (methodFn_instance_var, actual_method->src);
  }
  if ((nParams_instance_var = 
       __ctalkGetInstanceVariable (rcvr_obj, "nParams", TRUE)) == NULL)
    return ERROR;
  sprintf (s, "%d", actual_method ->n_params);
  __ctalkSetObjectValueVar (nParams_instance_var, s);
  if ((varargs_instance_var = 
       __ctalkGetInstanceVariable (rcvr_obj, "varargs", TRUE)) == NULL)
    return ERROR;
  sprintf (s, "%d", actual_method ->varargs);
  __ctalkSetObjectValueVar (varargs_instance_var, s);
  if ((nArgs_instance_var = 
       __ctalkGetInstanceVariable (rcvr_obj, "nArgs", TRUE)) == NULL)
    return ERROR;
  sprintf (s, "%d", actual_method -> n_args);
  __ctalkSetObjectValueVar (nArgs_instance_var, s);
  if ((argFrameTop_instance_var = 
       __ctalkGetInstanceVariable (rcvr_obj, "argFrameTop", TRUE)) == NULL)
    return ERROR;
  sprintf (s, "%d", actual_method -> arg_frame_top);
  __ctalkSetObjectValueVar (argFrameTop_instance_var, s);
  if ((rcvrFrameTop_instance_var = 
       __ctalkGetInstanceVariable (rcvr_obj, "rcvrFrameTop", TRUE)) == NULL)
    return ERROR;
  sprintf (s, "%d", actual_method -> rcvr_frame_top);
  __ctalkSetObjectValueVar (rcvrFrameTop_instance_var, s);
  return SUCCESS;
}

int __ctalkDefinedInstanceMethodObject (OBJECT *rcvr, char *classname, char *name) {
  OBJECT *rcvr_obj, *class_obj;
  METHOD *defined_method, *__m;

  rcvr_obj = (IS_VALUE_INSTANCE_VAR(rcvr) ? rcvr -> __o_p_obj : rcvr);
  if ((class_obj = __ctalkGetClass (classname)) == NULL) {
    _warning ("__ctalkDefinedMethodObject: Undefined class %s.\n",
	      classname);
    return ERROR;
  }
  for (defined_method = NULL, __m = class_obj -> instance_methods; 
       __m && !defined_method; __m = __m -> next) 
    if (!strcmp (__m -> name, name)) 
      defined_method = __m;
  if (!defined_method) {
    _warning ("__ctalkDefinedInstanceMethodObject: Undefined method %s (class %s).\n",
	      name, classname);
    return ERROR;
  }
  return __method_object_init (rcvr_obj, class_obj, defined_method);
}

int __ctalkDefinedClassMethodObject (OBJECT *rcvr, char *classname, char *name) {
  OBJECT *rcvr_obj, *class_obj;
  METHOD *defined_method, *__m;

  rcvr_obj = (IS_VALUE_INSTANCE_VAR(rcvr) ? rcvr -> __o_p_obj : rcvr);
  if ((class_obj = __ctalkGetClass (classname)) == NULL) {
    _warning ("__ctalkDefinedMethodObject: Undefined class %s.\n",
	      classname);
    return ERROR;
  }
  for (defined_method = NULL, __m = class_obj -> class_methods; 
       __m && !defined_method; __m = __m -> next) 
    if (!strcmp (__m -> name, name)) 
      defined_method = __m;
  if (!defined_method) {
    _warning ("__ctalkDefinedClassMethodObject: Undefined method %s (class %s).\n",
	      name, classname);
    return ERROR;
  }
  return __method_object_init (rcvr_obj, class_obj, defined_method);
}

static ARG *method_object_tmp_args[MAXARGS] = {NULL,};
static int method_object_tmp_arg_ptr = 0;
static METHOD *this_method_tmp;
int method_object_message_call = FALSE;
void save_method_object_args (METHOD *m) {
  if (!strcmp (m->name, "methodObjectMessage") && 
      (m->n_args > 1)) {
    for (method_object_tmp_arg_ptr = 0; 
	 m -> args[method_object_tmp_arg_ptr];
	 method_object_tmp_arg_ptr++) {
      method_object_tmp_args[method_object_tmp_arg_ptr] = 
	__ctalkCreateArgEntryInit (m->args[method_object_tmp_arg_ptr] -> obj);
      __ctalkDeleteArgEntry (m->args[method_object_tmp_arg_ptr]);
      m -> args[method_object_tmp_arg_ptr] = NULL;
    }
    m -> args[0] = method_object_tmp_args[method_object_tmp_arg_ptr - 1];
  }
}

void restore_method_object_args (METHOD *m) {
  int i;
  if (method_object_tmp_arg_ptr) {
    for (i = 0; i < method_object_tmp_arg_ptr; i++) {
      m -> args[i] = method_object_tmp_args[i];
      method_object_tmp_args[i] = NULL;
    }
    m -> n_args = method_object_tmp_arg_ptr;
    method_object_tmp_arg_ptr = 0;
  }
}

int method_object_args_ok (METHOD *m) {
  if (((m -> n_args == method_object_tmp_arg_ptr)  &&
       method_object_message_call) ||
      (m -> n_args == 1))
    return TRUE;
  else
    return FALSE;
}

#define __LIST_HEAD(__o) ((__o)->instancevars->next)

int __ctalkMethodObjectMessage (OBJECT *rcvr, OBJECT *method_instance) {
  OBJECT *rcvr_obj, *method_obj, *methodArgs_instance_var, 
    *nArgs_instance_var, *nParams_instance_var, *varargs_instance_var,
    *methodName_instance_var, *methodFn_instance_var,
    *t, *t_prev, *arg_ref, *rcvr_obj_tmp,
    *(*fn)();
  int n_args, n_params, varargs, n_th_arg, arg_refs_orig[MAXARGS];
  METHOD *target_method;
  METHOD *this_method;
  if ((this_method = __ctalkRtGetMethod ()) == NULL) {
    fprintf (stderr, "__ctalkMethodObjectMessage: can't find calling method.\n");
    return ERROR;
  }
  this_method_tmp = this_method;
  save_method_object_args (this_method);
  rcvr_obj = IS_VALUE_INSTANCE_VAR(rcvr) ? rcvr->__o_p_obj : rcvr;
  method_obj = IS_VALUE_INSTANCE_VAR(method_instance) ? 
    method_instance->__o_p_obj :
    method_instance;
  if ((methodName_instance_var = 
       __ctalkGetInstanceVariable (method_obj, "methodName", TRUE)) == NULL)
    return ERROR;
  if ((methodArgs_instance_var = 
       __ctalkGetInstanceVariable (method_obj, "methodArgs", TRUE)) == NULL)
    return ERROR;
  if ((nArgs_instance_var = 
       __ctalkGetInstanceVariable (method_obj, "nArgs", TRUE)) == NULL)
    return ERROR;
  if ((nParams_instance_var = 
       __ctalkGetInstanceVariable (method_obj, "nParams", TRUE)) == NULL)
    return ERROR;
  if ((varargs_instance_var = 
       __ctalkGetInstanceVariable (method_obj, "varargs", TRUE)) == NULL)
    return ERROR;
  if ((methodFn_instance_var = 
       __ctalkGetInstanceVariable (method_obj, "methodFn", TRUE)) == NULL)
    return ERROR;
  n_args = atoi(nArgs_instance_var->instancevars->__o_value);
  n_params = atoi(nParams_instance_var->instancevars->__o_value);
  varargs = atoi(varargs_instance_var->instancevars->__o_value);
  if (n_args != n_params) {
    if (!varargs) {
      char ebuf[MAXMSG];
      sprintf (ebuf, "Wrong number of arguments to %s", method_obj->__o_name);
      __ctalkExceptionInternal (NULL, invalid_operand_x, ebuf, 0);
    }
  } else {
    rcvr_obj_tmp = rcvr_obj;
    if ((target_method = __ctalkFindInstanceMethodByName 
	 (&rcvr_obj_tmp, methodName_instance_var->instancevars->__o_value,
	  TRUE)) == NULL)
      return ERROR;
    if ((methodArgs_instance_var = 
	 __ctalkGetInstanceVariable (method_obj, "methodArgs", TRUE)) == NULL)
      return ERROR;
    target_method->n_args = 0;
    for (t = __LIST_HEAD(methodArgs_instance_var); t; t = t -> next) {
      if ((arg_ref = obj_ref_str (t -> instancevars->__o_value)) != NULL) {
	arg_refs_orig[target_method->n_args] = arg_ref -> nrefs;
	__add_arg_object_entry_frame (target_method, arg_ref);
	__ctalk_arg_push_ref (arg_ref);
 	__ctalkSetObjectScope (arg_ref, 
 			       arg_ref -> scope | VAR_REF_OBJECT);
      }
    }
    sscanf (methodFn_instance_var->instancevars->__o_value,
	    "0x%p", &fn);
    method_object_message_call = TRUE;
    __ctalk_method_from_object (rcvr_obj, fn);
    restore_method_object_args(this_method);
    method_object_message_call = FALSE;
    for (t = __LIST_HEAD(methodArgs_instance_var); 
	 t && t -> next; t = t -> next)
      ;
    for (n_th_arg = 0; n_th_arg < n_args; n_th_arg++) {
      if (t) {
	OBJECT *__arg_ref_tmp, *__arg_object_tmp;
	int nrefs_tmp;
	__arg_ref_tmp = obj_ref_str (t -> instancevars->__o_value);
	__arg_object_tmp = __ctalk_arg_pop ();
	__ctalk_arg_push (__arg_object_tmp);
	nrefs_tmp = __ctalk_arg_cleanup ();
	if ((__arg_ref_tmp == __arg_object_tmp) && !nrefs_tmp) {
	  *t->__o_value = '\0';
	  *t->instancevars->__o_value = '\0';
	}

	t_prev = t -> prev;
	/*
	 *  Arg's ref count is incremented by List : pushItemRef.
	 */
	if ((arg_ref = obj_ref_str (t -> instancevars->__o_value)) != NULL) {
	  __objRefCntSet(OBJREF(arg_ref), arg_refs_orig[n_th_arg]);
	  __ctalkSetObjectScope (arg_ref, arg_ref->scope & ~VAR_REF_OBJECT);
	}
	/* Don't adjust arg's reference count again when deleting from argument 
	   list. */
	t->instancevars->__o_value[0] = '\0';
	__ctalkDeleteObject(t);
	t = t_prev;
	if (t) t -> next = NULL;
      }
    }
    methodArgs_instance_var->instancevars->next = NULL;
    __ctalkSetObjectValue (nArgs_instance_var->instancevars, "0");
  }
return SUCCESS;
}

OBJECT *__ctalk_method_from_object (OBJECT *rcvr_obj, OBJECT *(*__cfunc)()) {
  OBJECT *__method_rcvr_obj_, *__class_obj_, *__result_obj_ = NULL;
  METHOD *method;
  int i;

  __class_obj_ = rcvr_obj->__o_class;
  __method_rcvr_obj_ = rcvr_obj;

  if (((method = 
       __ctalkFindInstanceMethodByFn (&__method_rcvr_obj_, __cfunc, FALSE))
      == NULL) &&
      ((method = 
	__ctalkFindClassMethodByFn (&__method_rcvr_obj_, __cfunc, FALSE))
       == NULL)) {
    if (rcvr_obj -> instancevars) {
      __method_rcvr_obj_ = rcvr_obj -> instancevars;
      if (((method = 
	    __ctalkFindInstanceMethodByFn (&__method_rcvr_obj_, 
					   __cfunc, FALSE))
	   == NULL) &&
	  ((method = 
	    __ctalkFindClassMethodByFn (&__method_rcvr_obj_, 
					__cfunc, FALSE))
	   == NULL)) {
	/*
	 *  Experimental - if the receiver is the member of a collection, 
	 *  then look for the method in the parent object's class.  Then, 
	 *  if the parent object is an instance or class variable, look 
	 *  for the method in the parent object's value class.
	 */
	if (rcvr_obj -> __o_p_obj && 
	    IS_OBJECT (rcvr_obj -> __o_p_obj)) {
	  __method_rcvr_obj_ = rcvr_obj -> __o_p_obj;
	  if (((method = 
		__ctalkFindInstanceMethodByFn (&__method_rcvr_obj_, 
					       __cfunc, FALSE)) == NULL) &&
	      ((method = 
		__ctalkFindClassMethodByFn (&__method_rcvr_obj_, 
					    __cfunc, FALSE))  == NULL)) {
	    OBJECT *p_value_obj;
	    if ((p_value_obj = 
		 __ctalkGetInstanceVariable (rcvr_obj -> __o_p_obj,
					     "value", FALSE)) != NULL) {
	      __method_rcvr_obj_ = p_value_obj;
	      if (((method = 
		    __ctalkFindInstanceMethodByFn (&__method_rcvr_obj_, 
						   __cfunc, FALSE)) == NULL) &&
		  ((method = 
		    __ctalkFindClassMethodByFn (&__method_rcvr_obj_, 
						__cfunc, FALSE))  == NULL)) {
		_warning ("__ctalk_method: Undefined method.\n");
		return NULL;
	      } 
	    }
	  }  else {
	    _warning ("__ctalk_method: Undefined method.\n");
	    return NULL;
	  }
	} 
      }
    }
  }

  if ((method -> n_args != method -> n_params) &&
      (method -> varargs == 0)) {
    if (rcvr_obj -> __o_superclass == 
	method -> rcvr_class_obj) {
      METHOD *__calling_method;
      if (((__calling_method = __ctalkRtGetMethod ()) != NULL) &&
	  !strcmp (method -> name, __calling_method -> name) &&
	  (__calling_method -> n_args > __calling_method -> n_params)) {
	OBJECT *__a;
	int __n_th_param;
	for (__n_th_param = 0; __n_th_param < method -> n_params; 
	     __n_th_param++) {
	  __a=__calling_method->args[__calling_method->n_args-1]->obj;
	  __ctalkDeleteArgEntry
	    (__calling_method->args[__calling_method->n_args-1]);
	  __calling_method -> args[__calling_method -> n_args - 1] = NULL;
	  __calling_method -> n_args -= 1;
	  __objRefCntDec (OBJREF(__a));
	  __add_arg_object_entry_frame (method, __a);
	}
      } else {
	__ctalkExceptionInternal (NULL, 
				  wrong_number_of_arguments_x, 
				  method -> name,0);
	__ctalkHandleRunTimeExceptionInternal ();
	return null_result_obj (method, ARG_VAR);
      }
    } else {
      /*
       *  Handle exception immediately in case called by a top-level
       *  function.
       */
      __ctalkExceptionInternal (NULL, wrong_number_of_arguments_x, method -> name,0);
      __ctalkHandleRunTimeExceptionInternal ();
      return null_result_obj (method, ARG_VAR);
    }
  }

  method -> arg_frame_top = __ctalkGetArgPtr () + 1;
  method -> rcvr_frame_top = __ctalkGetReceiverPtr () + 1;

  __save_rt_info (rcvr_obj, __class_obj_, NULL, method, __cfunc, False);

  __ctalk_receiver_push_ref (rcvr_obj);

  __result_obj_ = (__cfunc)();

  __ctalk_receiver_pop_deref ();

  __ctalk_methodReturn (rcvr_obj, method, __result_obj_);

  if (IS_CONSTRUCTOR (method)) {
    OBJECT *__arg_object;
    VARENTRY *t_local, *v_next;
    int call_level;
    /*
     *  The interpreter adds a __ctalk_set_local () call in the
     *  caller's init block, so it isn't needed here.
     *
     *  The object must be removed from the method, so it 
     *  isn't bollixed the next time the method is called.
     *
     *  If the argument is still on the stack, remove it.
     */
    if (__result_obj_ && IS_OBJECT(__result_obj_))
      __ctalk_dictionary_add (__result_obj_);

    __arg_object = __ctalk_arg_pop ();
    __ctalk_arg_push (__arg_object);
    if (__result_obj_ == __arg_object) {
      if (!strcmp (method -> rcvr_class_obj -> __o_name, OBJECT_CLASSNAME)) {
	__ctalk_arg_pop ();
      }
      method -> args[method -> n_args - 1] = NULL;
      --(method -> n_args);
    }

    for (t_local = method -> local_objects.vars; 
	 t_local; t_local = t_local -> next) {
      if (t_local -> var_object == __result_obj_) {
	if (__result_obj_ == method -> local_objects.vars->var_object) {
	  v_next = method -> local_objects.vars -> next;
	  delete_varentry (method -> local_objects.vars);
	  method -> local_objects.vars = t_local = v_next;
	} else {
	  if (t_local -> prev) t_local -> prev -> next = t_local -> next;
	  if (t_local -> next) t_local -> next -> prev = t_local -> prev;
	  delete_varentry (t_local);
	}
      }
      if (!t_local) 
	break;
    }

    /*
     *  If the calling method is also a constructor, then 
     *  set the result object's class and superclass to 
     *  the calling method.  Also set the value variable's
     *  class and superclass.
     *
     *  The caller is two call frames up from the call 
     *  stack pointer.
     *
     *  If the caller is a _rt_fn, then skip.
     */
    if ((call_level = __get_call_stack_ptr ()) < (MAXARGS - 1) &&
	(__call_stack[call_level+2] -> _rt_fn == NULL)) {
      METHOD *calling_method;
      OBJECT *calling_rcvr_class_obj;
      calling_rcvr_class_obj = 
	__call_stack[call_level + 2] -> rcvr_class_obj;
      if ((calling_method = 
	   __ctalkGetInstanceMethodByFn (calling_rcvr_class_obj, 
					 __call_stack[call_level+2]->method_fn,
					 FALSE)) != NULL) {
	if (IS_CONSTRUCTOR(calling_method)) {
	  OBJECT *result_object_value_var;
	  result_object_value_var = __result_obj_ -> instancevars;
	  strcpy (__result_obj_ -> __o_classname, 
		  calling_rcvr_class_obj -> __o_name);
	  strcpy (__result_obj_ -> __o_superclassname, 
		  calling_rcvr_class_obj -> __o_superclassname);
	  __result_obj_->__o_class = calling_rcvr_class_obj -> __o_class;
	  __result_obj_->__o_superclass = 
	    calling_rcvr_class_obj -> __o_superclass;
	}
      }    
    }
  }

  if (method -> varargs) {
    int arg_frame, n_args_tmp;
    arg_frame = method->args[method->n_args-1]->call_stack_frame;
    n_args_tmp = method -> n_args;
    for (i = 0; i < method -> n_args; i++) {
      if ((IS_ARG(method->args[i])) && 
	  (method->args[i]->call_stack_frame == arg_frame)) {
	__ctalkDeleteArgEntry (method->args[i]); 
	method -> args[i] = NULL;
	n_args_tmp--;
      }
    }
    method -> n_args = n_args_tmp;
  } else {
    for (i = 0; i < method -> n_args; i++) {
      if (IS_ARG(method -> args[i]))
	__ctalkDeleteArgEntry (method->args[i]);
    }
    method -> n_args = 0;
  }

  __restore_rt_info ();

  return __result_obj_;
}

